# Scotty TNG - ping/snmp based realtime network monitoring
#
# $Id: chart.rb 113 2008-10-02 07:56:09Z liske $
#
# Authors:
#   Thomas Liske <liske@ibh.de>
#   Andre Ellguth <ellguth@ibh.de>
#
# Copyright Holder:
#   2008 (C) IBH IT-Service GmbH [http://www.ibh.de/]
#
# License:
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this package; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
#

class Chart
  attr_reader :opts, :img

  CWidth = 60
  CHeight = 35
  CFoot = 10

  @@charts = Hash.new
  @@tooltips = Hash.new

  def initialize(nname, sname, opts)
    @nname = nname
    @node = Node.getNode(nname)
    @sname = sname
    @type = sname.match(/^\w+/)[0]
    @status = [nil, "ok", nil]

    @vcharts = Array.new
    @notid = 0

    @img = Cairo::ImageSurface.new(Cairo::FORMAT_ARGB32, CWidth, CHeight + CFoot)
    @ct = Cairo::Context.new(@img)

    @opts = opts
    unless @opts
      Logger.logtxt("Could not get chart: #{$!}")
      return
    end
    [:title, :descr].each{|k|
	@opts[k] = Entities.encode(@opts[k]) if @opts[k]
    }

    @@charts[[@nname, @sname]] = self

    @data = Array.new(@opts[:lines])
    @current = Array.new(@opts[:lines], 0)
    @show_current = Array.new(@opts[:lines], true)
    @last = Array.new(@opts[:lines], 0)
    @sum = Array.new(@opts[:lines], 0)
    @max = Array.new(@opts[:lines], 0)

    0.upto(@opts[:lines] - 1){ |l|
        @data[l] = Array.new
    }

    # Read nagios status
    stat = [Time.now.to_i, "ok", nil]
    status = Client.getTS.read_all([:SensorStatus, @nname, @sname, nil])
    status.each{ |s|
      stat = s[3]
    }

    # Read current status
    setStatus(stat, true)
  end

  def drawUpdate
    draw_chart_data(@ct)
  end

  # Assigns a +VChart+ to a +Chart+ instance.
  def add(vchart)
    @vcharts.push(vchart)
    vchart.setStatus(@status[1])
  end

  # Add the given +VChart+ to a named +Chart+ instance.
  def self.add(nname, sname, vchart, opts=nil)
    if @@charts[[nname, sname]] then
	chart = @@charts[[nname, sname]]
    else
	chart = Chart.new(nname, sname, opts)
    end

    chart.add(vchart) if vchart

    return chart
  end

  def pushVal(dval)
    0.upto(@opts[:lines] - 1) do |l|
      @show_current[l] = (dval[l] != nil)
      dval[l] = 0 unless @show_current[l]

      case @opts[:type][l]
      when "c"
        if @current[l] - dval[l] > 0 then
          val = (@current[l] - dval[l]) / (Time.now.to_f - @last)
        else
          val = 0
        end
      else
        val = dval[l]
      end
      
      @current[l] = dval[l]

      if @data[l].size > CWidth then
        @sum[l] -= @data[l].shift

	@max[l] = 0
	@data[l].each{|d|
	    @max[l] = d * 1.2 if d > @max[l]
	}
      end

      @data[l].push(val)
      @sum[l] += val
      
      @max[l] = val * 1.2 if @max[l] == nil or @max[l] < val
    end
      
    @last = Time.now.to_f
    
    @vcharts.each{|chart|
	chart.redraw
    }

    updTooltip
  end

  # Draws the content of a chart
  def draw_frame(ct)
    ct.set_source_rgb(0.9, 0.9, 0.9)
    ct.rectangle(0, CHeight, CWidth, CFoot)
    ct.fill_preserve
    ct.stroke

    ct.set_source_rgb(0, 0, 0)
    ct.move_to(1, 9)
    ct.show_text(@opts[:title]) if @opts
    ct.stroke

    Client.set_ct_scol(ct, Status.text_color(@status[1]))
    ct.rectangle(0, 0, CWidth, CHeight)
    ct.stroke
  end

  def draw_chart_data(ct)
    height = CHeight
    width = CWidth

    ct.line_width = 1
    ct.set_font_size(9);

    # Clear chart
    ct.set_source_rgb(1, 1, 1)
    ct.rectangle(0, 0, width, height)
    ct.fill_preserve
    ct.stroke

    # Warn about unkown sensors.
    unless @opts then
	draw_frame(ct)

	ct.set_font_size(8);
	ct.set_source_rgb(0.8, 0, 0)

	ct.move_to(1, Chart::CHeight + Chart::CFoot - 2)
	ct.show_text("Sensor N/A!")

	return
    end


    # Get maximum values (dynamic for scaling)
    # The max values are grouped by opts[:max]
    totalmax = Hash.new
    0.upto(@opts[:lines] - 1) do |l|
	totalmax[@opts[:max][l]] = @max[l] if totalmax[@opts[:max][l]] == nil or @max[l] > totalmax[@opts[:max][l]]
    end

    # Draw the lines
    hasdata = 0
    0.upto(@opts[:lines] - 1) do |l|
	next if @data[l].length == 0
	hasdata = 1
	next if @opts[:slegend][l] == nil
	Client.set_ct_scol(ct, @opts[:color][l])

	if @opts[:max][l] > 0 then
	    max = totalmax[@opts[:max][l]]
	else
	    max = @max[l]
	end
	max = 1 if max == 0

	first = 1
	x = 1
	@data[l].each{|p|
	    begin
		y = height - p * (height - 10) / max
		y = 1 if y < 1
	    rescue
		y = height - 1
	    end

	    if first == 1 then
		if l == 0 then
		    ct.move_to(x, height)
		    ct.line_to(x, y)
		else
		    ct.move_to(x, y)
		end
		first = 0
	    else
		ct.line_to(x, y)
	    end
	    x += 1
	}
	if l == 0 then
	    ct.line_to(x-1, height)
	    ct.fill
	end
	ct.stroke
    end

    # Warn about missing data
    if hasdata == 0 then
	draw_frame(ct)

	ct.set_font_size(8);
	ct.set_source_rgb(0.8, 0, 0)

	ct.move_to(1, Chart::CHeight + Chart::CFoot - 2)
	ct.show_text("No data!")

	return
    end

    ct.set_dash([2], 0)
    0.upto(@opts[:lines] - 1) do |l|
	next if @opts[:slegend][l] == nil

	begin
	    Client.set_ct_scol(ct, "dark" + @opts[:color][l])

	    if @opts[:max][l] > 0 then
		max = totalmax[@opts[:max][l]]
	    else
		max = @max[l]
	    end
	    max = 1 if max == 0

	    avg = @sum[l]/@data[l].size
	    y = height - avg * (height - 10) / max

	    ct.move_to(      0, y)
	    ct.line_to(width-1, y)
	    ct.stroke
	rescue
	    p $!
	end
    end
    ct.set_dash(nil, 0)

    # Draw frame
    draw_frame(ct)

    # Print current values
    ct.set_font_size(8);
    ct.move_to(1, CHeight + CFoot - 2)

    0.upto(@opts[:lines] - 1) do |l|
	if @opts[:slegend][l] != nil and
	   @show_current[l] then
	    Client.set_ct_scol(ct, "dark" + @opts[:color][l])
	    ct.show_text(Client.format_data(@opts[:format][l], @data[l].last) + " ")
	end
    end
    ct.stroke
  end

  def setStatus(nstatus, startup=false)
    if nstatus[1] != @status[1] or nstatus[2] != @status[2] or startup then
	if @notid > 0 then
	    Notify.terminate(@notid)
	    @notid = 0
	end

	if nstatus[1] != "ok" or not startup then
	    msg = "#{@opts[:title]} "
	    msg += "[#{@opts[:descr]}] " if @opts[:descr]
	    msg += "@ #{@nname} changed to #{nstatus[1].upcase}"
	    msg += ": #{nstatus[2]}" if nstatus[2]
	    Logger.logtxt(msg, false, nstatus[1], nstatus[0])
	    Chart.updSIcon(self, nstatus)
	    to = nstatus[1]
	    from = @status[1]

	    Client.getTrans.each{ |t|
		if ((t[0] == nil) or (t[0] == from)) and
		    ((t[1] == nil) or (t[1] == to)) and
		    ((t[2] == nil) or (t[2] == @type)) and
		    ((t[3] == nil) or (@node == nil) or (t[3] == @node.opts[:class])) then
		    if t[4] != nil then
			action = Client.getActs[t[4]]
			if action == nil then
			    Logger.logtxt("Could not find actions '#{action}' for state transition #{t[0]} => #{t[1]}!")
			else
			    @notid = Notify.notification(action[0], action[1], to, Status.text_color(t[1]), msg)
			end
		    end

		    break
		end
	    }
	end

	@status = nstatus
	Node.getNode(@nname).setStatus(@sname, @status) if Node.getNode(@nname)
	updTooltip

	@vcharts.each{|vchart|
	    vchart.setStatus(@status[1])
	}
    end
  end

  def self.updSIcon(chart, status)
    if status[1] == "ok" then
	@@tooltips.delete(chart)
    else
	@@tooltips[chart] = status
    end

    return unless Client.getSIcon

    tooltip = "Scotty TNG\n"

    stat = "ok"
    @@tooltips.each_pair{|k, s|
	tooltip += "\n#{k.opts[:title]}: #{s[2]}"

	stat = s[1] if Status.int(s[1]) > Status.int(stat)
    }

    Client.getSIcon.file = "res/rs" + Status.icon_color(stat) + ".png"
    Client.getSIcon.tooltip = tooltip
    Client.getSIcon.blinking = (stat == "critical")
  end

  def updTooltip

    tt  = "<b>#{@opts[:title]}"
    tt += " on #{@nname}" if @nname != ":SensorStack"
    tt += "</b>\n"
    tt += "<small>#{@opts[:descr]}</small>\n" if @opts[:descr] != nil

    ttmax = 6
    0.upto(@opts[:lines] - 1) do |l|
      ttmax = @opts[:legend][l].length if ttmax < @opts[:legend][l].length
    end

    tt += "<tt>" + sprintf("\n%#{ttmax}s: ", "Status") 
    tt += "<b><span foreground='" + Status.text_color(@status[1]) + "'>"
    tt += @status[1].upcase + "</span></b></tt>\n"
    tt += "<tt>" + sprintf("%#{ttmax}s: ", "Since") 
    tt += Time.at(@status[0]).strftime("%x %X") + "</tt>\n"
    tt += "<small>" + @status[2] + "</small>\n" if @status[2]
    
    0.upto(@opts[:lines] - 1) do |l|  
      next if @opts[:legend][l] == nil
      next unless @data[l].size > 0

      tt += "\n<tt><span foreground='" + @opts[:color][l] + "'>"
      tt += sprintf("%#{ttmax}s", @opts[:legend][l]) + ": "
      if @opts[:format][l] != 'b' then
	tt += Client.format_data(@opts[:format][l], @sum[l]/@data[l].size)
      else
	tt += Client.format_data('p', @sum[l]/@data[l].size*100)
      end
      tt += "</span></tt>"
    end

    @vcharts.each{|chart|
	chart.setTooltip(tt)
    }
  end

  def self.getChart(lname)
    @@charts[lname]
  end
end

class VChart
  attr_reader :x, :y, :view

  CWidth = 60
  CHeight = 35
  CFoot = 10

  @@tooltips = Hash.new

  def initialize(view, chart, layout)
    @view = view

    @nname = chart.attributes["node"]
    @sname = chart.attributes["sensor"]

    begin
	@x = chart.attributes["x"].to_i
	@y = chart.attributes["y"].to_i
    rescue
    end
    @visible = (@x != 0 and @y != 0)

    chartobj = Chart.add(@nname, @sname, self)

    @status = [nil, "ok", nil]
    @da = nil
    if visible? then
	@eb = Gtk::EventBox.new
	@eb.above_child = true
	@eb.visible_window = false
	@da = Gtk::DrawingArea.new
	@eb.add(@da)
	@da.set_size_request(CWidth, CHeight + CFoot)
	@da.signal_connect("expose_event") { |widget, arg0|
	  chartobj.drawUpdate

	  # Copy image to drawing area surface
	  c = widget.window.create_cairo_context
	  c.set_source(chartobj.img, 0, 0)
	  c.paint
	}

	@sx = chart.attributes["x"].to_i
	@sy = chart.attributes["y"].to_i

	layout.put(@eb, @x, @y)
	@eb.signal_connect("size_allocate") { |widget, alloc|
	    nx = @x - alloc.width/2
	    ny = @y - alloc.height/2

	  if @sx != nx or
	    @sy != ny then
	    @sx = nx
	    @sy = ny
	    layout.move(widget, nx, ny)
	  end
	}
	@menu = Gtk::Menu.new
	@mi_debug = Gtk::MenuItem.new("Debug")
	@mi_debug.signal_connect("activate") do
	     dialog = Gtk::Dialog.new("Chart Debug",
		nil, nil,
		[Gtk::Stock::OK, Gtk::Dialog::RESPONSE_NONE])
	    dialog.signal_connect('response') { dialog.destroy }
	    dialog.vbox.add(Gtk::Label.new("Node: #{@nname}\nSensor: #{@sname}\nCoords: #{@x}x#{@y}"))
	    dialog.show_all
	end
	@menu.append(@mi_debug)
	@menu.show_all

	@eb.signal_connect("button_press_event") do |widget, arg0|
	    @menu.popup(nil, nil, arg0.button, arg0.time) if arg0.button == 3
	end
	@da.show_all
    end
  end

  def visible?
    @visible
  end

  def setTooltip(tt)
    return unless visible?

    begin
	# Older GTK bindings lack tooltip_markup property
	@da.tooltip_markup = tt.chomp
    rescue
	# Gtk::DrawingArea has now window, use the Gtk::EventBox for tooltips
	Client.getTooltips.set_tip(@eb, tt.gsub(/<[^>]+>/, "").chomp, nil)
    end
  end

  def setStatus(status="ok")
    @view.setStatus("s:#{@nname}/#{@sname}", status)
  end

  def redraw
    @da.queue_draw if @da
  end
end
