# Scotty TNG - ping/snmp based realtime network monitoring
#
# $Id: if.rb 111 2008-10-02 07:05:55Z liske $
#
# Authors:
#   Thomas Liske <liske@ibh.de>
#   Andre Ellguth <ellguth@ibh.de>
#
# Copyright Holder:
#   2008 (C) IBH IT-Service GmbH [http://www.ibh.de/]
#
# License:
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this package; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
#

module Sensors

  # This class implements a SNMP interface +Sensor+.
  class IF < Sensor
    @@sensors = Array.new
    
    attr_reader :ifInOctets_OID, :ifOutOctets_OID,
      :ifInErrors_OID, :ifOutErrors_OID,
      :ifInDiscards_OID, :ifOutDiscards_OID,
      :ifAdminStatus_OID, :ifOperStatus_OID,
      :rlname
    
    attr_accessor :mon_speed, :mon_in, :mon_out, :mon_errs,
      :mon_disc, :mon_aup, :mon_oup, :mon_reachable

    def initialize(node, xmlsensor)
      @@sensors.push(self)
      
      id = xmlsensor.attributes["idx"].to_i.to_s
      @rlname = "if[#{id}]"
      @recs = Array.new
      @stamps = Array.new

      @mon_speed = 0
      @mon_in = 0
      @mon_out = 0
      @mon_errs = 0
      @mon_disc = 0
      @mon_aup = 0
      @mon_oup = 0
      @mon_reachable = 0

      ifName_OID = "1.3.6.1.2.1.31.1.1.1.1." + id
      ifSpeed_OID = "1.3.6.1.2.1.2.2.1.5." + id
      ifHighSpeed_OID = "1.3.6.1.2.1.31.1.1.1.15." + id
      ifAlias_OID = "1.3.6.1.2.1.31.1.1.1.18." + id

      ifHCInOctets_OID = "1.3.6.1.2.1.31.1.1.1.6." + id
      ifHCOutOctets_OID = "1.3.6.1.2.1.31.1.1.1.10." + id
      @ifInOctets_OID = "1.3.6.1.2.1.2.2.1.10." + id
      @ifOutOctets_OID = "1.3.6.1.2.1.2.2.1.16." + id
      @ifInErrors_OID = "1.3.6.1.2.1.2.2.1.14." + id
      @ifOutErrors_OID = "1.3.6.1.2.1.2.2.1.20." + id
      @ifInDiscards_OID = "1.3.6.1.2.1.2.2.1.13." + id
      @ifOutDiscards_OID = "1.3.6.1.2.1.2.2.1.19." + id
      @ifAdminStatus_OID = "1.3.6.1.2.1.2.2.1.7." + id
      @ifOperStatus_OID = "1.3.6.1.2.1.2.2.1.8." + id

      super(node, xmlsensor, @rlname)

      begin
	response = node.snmp.get([ifName_OID, ifSpeed_OID, ifHighSpeed_OID, ifAlias_OID,
				   ifHCInOctets_OID, ifHCOutOctets_OID])
      rescue
	puts "\nERROR: Could not query SNMP: #{$!}"
	exit 1
      end

      name = nil
      @mon_speed = nil
      descr = "Interface ##{id}"

      response.each_varbind do |vb|
	begin
	    case vb.name.to_s
	    when ifName_OID
		name = vb.value.to_s.sub(/\-aal5 layer$/, '')
	    when ifSpeed_OID
		@mon_speed = vb.value.to_i
	    when ifHighSpeed_OID
		@mon_speed = vb.value.to_i * 1000000 if @mon_speed < vb.value.to_i
	    when ifAlias_OID
		descr = "Interface ##{id}\n#{vb.value.to_s.chomp}" if vb.value.to_s.chomp.length > 0
	    when ifHCInOctets_OID
		@ifInOctets_OID = ifHCInOctets_OID if vb.value.to_s != "noSuchInstance" && vb.value.to_s != "noSuchObject"
	    when ifHCOutOctets_OID
		@ifOutOctets_OID = ifHCOutOctets_OID if vb.value.to_s != "noSuchInstance" && vb.value.to_s != "noSuchObject"
	    end
	rescue
	    puts "Error retrieving #{vb.name.to_s}: #{$!}"
	end
      end
      
      RingList.new(node, self, @rlname , {
		     :lines => 6,
		     :title => name,
		     :descr => descr,
		     :max => [@mon_speed, @mon_speed, 0, 0, 1, 1],
		     :format => ["r", "r", "c", "c", "b", "b"],
		     :type => ["v", "v", "c", "c", "v", "v"],
		     :color => ["Green", "Blue", "Red", "Purple", "Brown", "Orange"],
		     :legend => ["In (bit/s)", "Out (bit/s)", "Errors (1/s)", "Discards (1/s)", "Admin up", "Oper up"],
		     :slegend => ["i", "o", "e", "d", nil, nil],
		   })
      self.add_expr("speed") { @mon_speed }
      self.add_expr("in") { @mon_in }
      self.add_expr("out") { @mon_out }
      self.add_expr("err") { @mon_errs }
      self.add_expr("disc") { @mon_disc }
      self.add_expr("aup") { @mon_aup }
      self.add_expr("oup") { @mon_oup }
      self.add_expr("reachable") { @mon_reachable }
    end

    def self.monitor()
      while true do
	@@sensors.each{ |s|
	    iocts = 0
	    oocts = 0
	    ierrs = 0
	    oerrs = 0
	    idisc = 0
	    odisc = 0
	    admin = 0
	    oper = 0
	    reachable = 0

	    begin
	      response = s.node.snmp.get([
					 s.ifInOctets_OID, s.ifOutOctets_OID,
					 s.ifInErrors_OID, s.ifOutErrors_OID,
					 s.ifInDiscards_OID, s.ifOutDiscards_OID,
					 s.ifAdminStatus_OID, s.ifOperStatus_OID,
				       ])
	      response.each_varbind do |vb|
	        vname = vb.name.to_s
	        case vb.name.to_s
	        when s.ifInOctets_OID
		  begin
		    iocts = s.getRate(vb.value.to_i, 0) * 8.0
		  rescue
		  end
	        when s.ifOutOctets_OID
		  begin
		    oocts = s.getRate(vb.value.to_i, 1) * 8.0
		  rescue
		  end
	        when s.ifInErrors_OID
		  begin
		    ierrs = vb.value.to_i
		  rescue
		  end
	        when s.ifOutErrors_OID
		  begin
		    oerrs = vb.value.to_i
		  rescue
		  end
	        when s.ifInDiscards_OID
		  begin
		    idisc = vb.value.to_i
		  rescue
		  end
	        when s.ifOutDiscards_OID
		  begin
		    odisc = vb.value.to_i
		  rescue
		  end
	        when s.ifAdminStatus_OID
		  begin
		    admin = vb.value.to_i
		  rescue
		  end
	        when s.ifOperStatus_OID
		  begin
		    oper = vb.value.to_i
		  rescue
		  end
	        end
	      end

	      reachable = 1
	    rescue
		print "#{s.node.name}/#{s.rlname}: "
		p $!
	    end

	    s.mon_in = iocts
	    s.mon_out = oocts
	    s.mon_errs = s.getRate(ierrs + oerrs, 2)
	    s.mon_disc = s.getRate(idisc + odisc, 3)
	    s.mon_oup = ifState(oper)
	    s.mon_aup = ifState(admin)
	    s.mon_reachable = reachable
	
	    s.check_mon
	
	    s.pushProbe([
						   s.mon_in,
						   s.mon_out,
						   (s.mon_errs == 0 ? nil : s.mon_errs),
						   (s.mon_disc == 0 ? nil : s.mon_disc),
						   s.mon_aup,
						   s.mon_oup,
						 ])
	}
	
	sleep 5
      end
    end

    def self.ifState(state)
	return 1 if state == 1
	return 0
    end

    def getRate(val, idx)
      old = @recs[idx].to_i
      last = @stamps[idx].to_i
      
      @recs[idx] = val
      @stamps[idx] = Time.now.to_i
      
      f = @stamps[idx] - last
      
      # Startup
      return 0 if last == 0 or f == 0
      
      # Wrapping
      return val / f if val < old
      
      return (val - old) / f
    end
  end
end
